@@ -13,7 +13,6 @@
 #    limitations under the License.
 #
 require 'faraday'
-
 require 'stringio'
 require 'addressable/uri'
 require 'signet'
@@ -57,6 +56,13 @@ module Signet
            instance_variable_set("@#{attr}", options[attr])
         end
       end
+ 
+      # Constant time string comparison.
+      def safe_equals?(a, b)
+        check = a.bytesize ^ b.bytesize
+        a.bytes.zip(b.bytes) { |x, y| check |= x ^ y.to_i }
+        check == 0
+      end
 
       ##
       # Determine if the supplied nonce/timestamp pair is valid by calling
@@ -285,7 +291,7 @@ module Signet
           client_credential_secret,
           nil
         )
-        if(computed_signature == auth_hash['oauth_signature'])
+        if safe_equals?(computed_signature, auth_hash['oauth_signature'])
           if(auth_hash.fetch('oauth_callback', 'oob').empty?)
             'oob'
           else
@@ -363,7 +369,7 @@ module Signet
           temporary_credential.secret
         )
 
-        if(computed_signature == auth_hash['oauth_signature'])
+        if safe_equals?(computed_signature, auth_hash['oauth_signature'])
           {:client_credential=>client_credential,
             :temporary_credential=>temporary_credential,
             :realm=>auth_hash['realm']
@@ -490,7 +496,7 @@ module Signet
           token_credential_secret
         )
 
-        if(computed_signature == auth_hash['oauth_signature'])
+        if safe_equals?(computed_signature, auth_hash['oauth_signature'])
           {:client_credential=>client_credential,
            :token_credential=>token_credential,
            :realm=>auth_hash['realm']
