@@ -19,6 +19,7 @@ require 'signet'
 require 'signet/errors'
 require 'signet/oauth_2'
 require 'jwt'
+require 'date'
 
 module Signet
   module OAuth2
@@ -88,14 +89,13 @@ module Signet
       #   )
       #
       # @see Signet::OAuth2::Client#update!
-      def initialize(options={})
+      def initialize options={}
         @authorization_uri    = nil
         @token_credential_uri = nil
         @client_id            = nil
         @client_secret        = nil
         @code                 = nil
         @expires_at           = nil
-        @expires_in           = nil
         @issued_at            = nil
         @issuer               = nil
         @password             = nil
@@ -104,6 +104,7 @@ module Signet
         @scope                = nil
         @state                = nil
         @username             = nil
+        @access_type          = nil
         self.update!(options)
       end
 
@@ -152,6 +153,8 @@ module Signet
       #     to be refreshed.
       #   - <code>:access_token</code> -
       #     The current access token for this client.
+      #   - <code>:access_type</code> -
+      #     The current access type parameter for #authorization_uri.
       #   - <code>:id_token</code> -
       #     The current ID token for this client.
       #   - <code>:extension_parameters</code> -
@@ -189,6 +192,7 @@ module Signet
         self.signing_key = options[:signing_key] if options.has_key?(:signing_key)
         self.extension_parameters = options[:extension_parameters] || {}
         self.additional_parameters = options[:additional_parameters] || {}
+        self.access_type = options.fetch(:access_type) { :offline }
         self.update_token!(options)
         return self
       end
@@ -233,6 +237,11 @@ module Signet
         # set, but this can be used to supply a more precise time.
         self.issued_at = options[:issued_at] if options.has_key?(:issued_at)
 
+        # Special case where we want expires_at to be relative to issued_at
+        if options.has_key?(:issued_at) && options.has_key?(:expires_in)
+          set_relative_expires_at options[:issued_at], options[:expires_in]
+        end
+
         self.access_token = options[:access_token] if options.has_key?(:access_token)
         self.refresh_token = options[:refresh_token] if options.has_key?(:refresh_token)
         self.id_token = options[:id_token] if options.has_key?(:id_token)
@@ -254,8 +263,8 @@ module Signet
         unless options[:response_type]
           options[:response_type] = :code
         end
-        unless options[:access_type]
-          options[:access_type] = :offline
+        if !options[:access_type] && access_type
+          options[:access_type] = access_type
         end
         options[:client_id] ||= self.client_id
         options[:redirect_uri] ||= self.redirect_uri
@@ -325,6 +334,23 @@ module Signet
         end
       end
 
+      ##
+      # Returns the current access type parameter for #authorization_uri.
+      #
+      # @return [String, Symbol] The current access type.
+      def access_type
+        return @access_type
+      end
+
+      ##
+      # Sets the current access type parameter for #authorization_uri.
+      #
+      # @param [String, Symbol] new_access_type
+      #   The current access type.
+      def access_type=(new_access_type)
+        @access_type = new_access_type
+      end
+
       ##
       # Returns the client identifier for this client.
       #
@@ -561,7 +587,7 @@ module Signet
       # Returns the number of seconds assertions are valid for
       # Used only by the assertion grant type.
       #
-      # @return [Fixnum] Assertion expiry, in seconds
+      # @return [Integer] Assertion expiry, in seconds
       def expiry
         return @expiry
       end
@@ -570,7 +596,7 @@ module Signet
       # Sets the number of seconds assertions are valid for
       # Used only by the assertion grant type.
       #
-      # @param [Fixnum, String] new_expiry
+      # @param [Integer, String] new_expiry
       #   Assertion expiry, in seconds
       def expiry=(new_expiry)
         @expiry = new_expiry ? new_expiry.to_i : nil
@@ -708,8 +734,10 @@ module Signet
       #   omitted.
       #
       # @return [String] The decoded ID token.
-      def decoded_id_token(public_key=nil, options = {})
-        payload, _header = JWT.decode(self.id_token, public_key, !!public_key, options)
+      def decoded_id_token public_key=nil, options = {}, &keyfinder
+        options[:algorithm] ||= signing_algorithm
+        verify = !!(public_key || keyfinder)
+        payload, _header = JWT.decode(self.id_token, public_key, verify, options, &keyfinder)
         if !payload.has_key?('aud')
           raise Signet::UnsafeOperationError, 'No ID token audience declared.'
         elsif payload['aud'] != self.client_id
@@ -721,39 +749,46 @@ module Signet
 
       ##
       # Returns the lifetime of the access token in seconds.
+      # Returns nil if the token does not expire.
       #
-      # @return [Fixnum] The access token lifetime.
+      # @return [Integer, nil] The access token lifetime.
       def expires_in
-        return @expires_in
+        if @expires_at.nil? || @issued_at.nil?
+          nil
+        else
+          (@expires_at - @issued_at).to_i
+        end
       end
 
       ##
-      # Sets the lifetime of the access token in seconds.  Resets the issued
-      # timestamp.
+      # Sets the lifetime of the access token in seconds.  Resets the issued_at
+      # timestamp. Nil values will be treated as though the token does
+      # not expire.
       #
-      # @param [String, Fixnum] new_expires_in
+      # @param [String, Integer, nil] new_expires_in
       #   The access token lifetime.
-      def expires_in=(new_expires_in)
-        if new_expires_in != nil
-          @expires_in = new_expires_in.to_i
+      def expires_in= new_expires_in
+        if !new_expires_in.nil?
           @issued_at = Time.now
+          @expires_at = @issued_at + new_expires_in.to_i
         else
-          @expires_in, @issued_at, @expires_at = nil, nil, nil
+          @expires_at = nil
+          @issued_at = nil
         end
       end
 
       ##
       # Returns the timestamp the access token was issued at.
       #
-      # @return [Time] The access token issuance time.
+      # @return [Time, nil] The access token issuance time.
       def issued_at
-        return @issued_at
+        @issued_at
       end
 
       ##
       # Sets the timestamp the access token was issued at.
       #
-      # @param [String,Fixnum,Time] new_issued_at
+      # @param [String,Integer,Time] new_issued_at
       #    The access token issuance time.
       def issued_at=(new_issued_at)
         @issued_at = normalize_timestamp(new_issued_at)
@@ -761,29 +796,26 @@ module Signet
 
       ##
       # Returns the timestamp the access token will expire at.
+      # Returns nil if the token does not expire.
       #
-      # @return [Time] The access token lifetime.
+      # @return [Time, nil] The access token lifetime.
       def expires_at
-        if @expires_at
-          @expires_at
-        elsif @issued_at && @expires_in
-          return @issued_at + @expires_in
-        else
-          return nil
-        end
+        @expires_at
       end
 
       ##
       # Limits the lifetime of the access token as number of seconds since
-      # the Epoch
-      # @param [String,Fixnum,Time] new_expires_at
-      #    The access token issuance time.
+      # the Epoch. Nil values will be treated as though the token does
+      # not expire.
+      # @param [String,Integer,Time, nil] new_expires_at
+      #    The access token expiration time.
       def expires_at=(new_expires_at)
-        @expires_at = normalize_timestamp(new_expires_at)
+        @expires_at = normalize_timestamp new_expires_at
       end
 
       ##
       # Returns true if the access token has expired.
+      # Returns false if the token has not expired or has an nil @expires_at.
       #
       # @return [TrueClass, FalseClass]
       #   The expiration state of the access token.
@@ -793,15 +825,15 @@ module Signet
 
       ##
       # Returns true if the access token has expired or expires within
-      # the next n seconds
+      # the next n seconds. Returns false for tokens with a nil @expires_at.
       #
-      # @param [Fixnum] sec
+      # @param [Integer] sec
       #  Max number of seconds from now where a token is still considered
       #  expired.
       # @return [TrueClass, FalseClass]
       #   The expiration state of the access token.
       def expires_within?(sec)
-        return self.expires_at.nil? || Time.now >= (self.expires_at - sec)
+        return self.expires_at != nil && Time.now >= (self.expires_at - sec)
       end
 
       ##
@@ -814,7 +846,7 @@ module Signet
         @password = nil
         @code = nil
         @issued_at = nil
-        @expires_in = nil
+        @expires_at = nil
       end
 
 
@@ -880,7 +912,7 @@ module Signet
       # @note A serialized client contains sensitive information. Persist or transmit with care.
       #
       # @return [String] A serialized JSON representation of the client.
-      def to_json
+      def to_json(*)
         return MultiJson.dump({
           'authorization_uri' => self.authorization_uri ? self.authorization_uri.to_s : nil,
           'token_credential_uri' => self.token_credential_uri ? self.token_credential_uri.to_s : nil,
@@ -940,14 +972,18 @@ module Signet
         end
         parameters['client_id'] = self.client_id unless self.client_id.nil?
         parameters['client_secret'] = self.client_secret unless self.client_secret.nil?
-        parameters['scope'] = options[:scope] if options[:scope]
+        if options[:scope]
+          parameters['scope'] = options[:scope]
+        elsif options[:use_configured_scope] && !self.scope.nil?
+          parameters['scope'] = self.scope
+        end
         additional = self.additional_parameters.merge(options[:additional_parameters] || {})
         additional.each { |k, v| parameters[k.to_s] = v }
         parameters
       end
 
       def fetch_access_token(options={})
-        if self.token_credential_uri == nil
+        if self.token_credential_uri.nil?
           raise ArgumentError, 'Missing token endpoint URI.'
         end
 
@@ -981,20 +1017,22 @@ module Signet
           raise ::Signet::AuthorizationError.new(
             message, :response => response
           )
+        elsif status.to_s[0] == "5"
+          message = 'Remote server error.'
+          if body.to_s.strip.length > 0
+            message += "  Server message:\n#{response.body.to_s.strip}"
+          end
+          raise ::Signet::RemoteServerError.new(message)
         else
           message = "Unexpected status code: #{response.status}."
           if body.to_s.strip.length > 0
             message += "  Server message:\n#{response.body.to_s.strip}"
           end
-          raise ::Signet::AuthorizationError.new(
-            message, :response => response
-          )
+          raise ::Signet::UnexpectedStatusError.new(message)
         end
       end
 
       def fetch_access_token!(options={})
-        options = deep_hash_normalize(options)
-
         token_hash = self.fetch_access_token(options)
         if token_hash
           # No-op for grant types other than `authorization_code`.
@@ -1010,8 +1048,6 @@ module Signet
       ##
       # Refresh the access token, if possible
       def refresh!(options={})
-        options = deep_hash_normalize(options)
-
         self.fetch_access_token!(options)
       end
 
@@ -1175,14 +1211,23 @@ module Signet
           nil
         when Time
           time
+        when DateTime
+          time.to_time
         when String
           Time.parse(time)
-        when Fixnum, Bignum
+        when Integer
           Time.at(time)
         else
           fail "Invalid time value #{time}"
         end
       end
+
+      def set_relative_expires_at(issued_at, expires_in)
+        self.issued_at = issued_at
+        # Using local expires_in because if self.expires_in is used, it returns
+        # the time left before the token expires
+        self.expires_at = self.issued_at + expires_in.to_i
+      end
     end
   end
 end
