@@ -15,6 +15,7 @@ require 'spec_helper'
 require 'signet/oauth_2/client'
 require 'openssl'
 require 'jwt'
+require 'date'
 
 conn = Faraday.default_connection
 
@@ -187,10 +188,10 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       @key = OpenSSL::PKey::RSA.new 2048
       @client = Signet::OAuth2::Client.new(
         :token_credential_uri =>
-          'https://accounts.google.com/o/oauth2/token',
+          'https://oauth2.googleapis.com/token',
         :scope => 'https://www.googleapis.com/auth/userinfo.profile',
         :issuer => 'app@example.com',
-        :audience => 'https://accounts.google.com/o/oauth2/token',
+        :audience => 'https://oauth2.googleapis.com/token',
         :signing_key => @key
       )
     end
@@ -199,10 +200,10 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       jwt = @client.to_jwt
       expect(jwt).not_to be_nil
 
-      claim, header = JWT.decode(jwt, @key.public_key, true)
+      claim, header = JWT.decode(jwt, @key.public_key, true, algorithm: 'RS256')
       expect(claim["iss"]).to eq 'app@example.com'
       expect(claim["scope"]).to eq 'https://www.googleapis.com/auth/userinfo.profile'
-      expect(claim["aud"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(claim["aud"]).to eq 'https://oauth2.googleapis.com/token'
     end
 
     it 'should generate valid JWTs for impersonation' do
@@ -210,11 +211,11 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       jwt = @client.to_jwt
       expect(jwt).not_to be_nil
 
-      claim, header = JWT.decode(jwt, @key.public_key, true)
+      claim, header = JWT.decode(jwt, @key.public_key, true, algorithm: 'RS256')
       expect(claim["iss"]).to eq 'app@example.com'
       expect(claim["prn"]).to eq 'user@example.com'
       expect(claim["scope"]).to eq 'https://www.googleapis.com/auth/userinfo.profile'
-      expect(claim["aud"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(claim["aud"]).to eq 'https://oauth2.googleapis.com/token'
     end
 
     it 'should generate valid JWTs for impersonation using deprecated person attribute' do
@@ -222,11 +223,11 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       jwt = @client.to_jwt
       expect(jwt).not_to be_nil
 
-      claim, header = JWT.decode(jwt, @key.public_key, true)
+      claim, header = JWT.decode(jwt, @key.public_key, true, algorithm: 'RS256')
       expect(claim["iss"]).to eq 'app@example.com'
       expect(claim["prn"]).to eq 'user@example.com'
       expect(claim["scope"]).to eq 'https://www.googleapis.com/auth/userinfo.profile'
-      expect(claim["aud"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(claim["aud"]).to eq 'https://oauth2.googleapis.com/token'
     end
 
     it 'should generate valid JWTs for impersonation using the sub attribute' do
@@ -234,11 +235,11 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       jwt = @client.to_jwt
       expect(jwt).not_to be_nil
 
-      claim, header = JWT.decode(jwt, @key.public_key, true)
+      claim, header = JWT.decode(jwt, @key.public_key, true, algorithm: 'RS256')
       expect(claim["iss"]).to eq 'app@example.com'
       expect(claim["sub"]).to eq 'user@example.com'
       expect(claim["scope"]).to eq 'https://www.googleapis.com/auth/userinfo.profile'
-      expect(claim["aud"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(claim["aud"]).to eq 'https://oauth2.googleapis.com/token'
     end
 
     it 'should generate a JSON representation of the client' do
@@ -247,18 +248,18 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       expect(json).not_to be_nil
 
       deserialized = MultiJson.load(json)
-      expect(deserialized["token_credential_uri"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(deserialized["token_credential_uri"]).to eq 'https://oauth2.googleapis.com/token'
       expect(deserialized["scope"]).to eq ['https://www.googleapis.com/auth/userinfo.profile']
       expect(deserialized["issuer"]).to eq 'app@example.com'
-      expect(deserialized["audience"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(deserialized["audience"]).to eq 'https://oauth2.googleapis.com/token'
       expect(deserialized["signing_key"]).to eq @key.to_s
     end
 
     it 'should send valid access token request' do
       stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-        stub.post('/o/oauth2/token') do |env|
+        stub.post('/token') do |env|
           params = Addressable::URI.form_unencode(env[:body])
-          claim, header = JWT.decode(params.assoc("assertion").last, @key.public_key)
+          claim, header = JWT.decode(params.assoc("assertion").last, @key.public_key, true, algorithm: 'RS256')
           expect(params.assoc("grant_type")).to eq ['grant_type','urn:ietf:params:oauth:grant-type:jwt-bearer']
           build_json_response({
             "access_token" => "1/abcdef1234567890",
@@ -282,10 +283,10 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       @key = 'my secret key'
       @client = Signet::OAuth2::Client.new(
         :token_credential_uri =>
-          'https://accounts.google.com/o/oauth2/token',
+          'https://oauth2.googleapis.com/token',
         :scope => 'https://www.googleapis.com/auth/userinfo.profile',
         :issuer => 'app@example.com',
-        :audience => 'https://accounts.google.com/o/oauth2/token',
+        :audience => 'https://oauth2.googleapis.com/token',
         :signing_key => @key
       )
     end
@@ -294,10 +295,10 @@ describe Signet::OAuth2::Client, 'configured for assertions profile' do
       jwt = @client.to_jwt
       expect(jwt).not_to be_nil
 
-      claim, header = JWT.decode(jwt, @key, true)
+      claim, header = JWT.decode(jwt, @key, true, algorithm: 'HS256')
       expect(claim["iss"]).to eq 'app@example.com'
       expect(claim["scope"]).to eq 'https://www.googleapis.com/auth/userinfo.profile'
-      expect(claim["aud"]).to eq 'https://accounts.google.com/o/oauth2/token'
+      expect(claim["aud"]).to eq 'https://oauth2.googleapis.com/token'
     end
   end
 end
@@ -308,7 +309,7 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
       :authorization_uri =>
         'https://accounts.google.com/o/oauth2/auth',
       :token_credential_uri =>
-        'https://accounts.google.com/o/oauth2/token',
+        'https://oauth2.googleapis.com/token',
       :scope => 'https://www.googleapis.com/auth/userinfo.profile'
     )
   end
@@ -367,6 +368,13 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     expect(request).to include('assertion' => 'PEFzc2VydGlvbiBJc3N1ZUluc3RhbnQ9IjIwMTEtMDU')
   end
 
+  it 'should include the scope in token request' do
+    @client.scope = ['https://www.googleapis.com/auth/userinfo.profile']
+
+    request = @client.generate_access_token_request(:use_configured_scope => true)
+    expect(request).to include('scope' => ['https://www.googleapis.com/auth/userinfo.profile'])
+  end
+
   it 'should allow the token to be updated' do
     issued_at = Time.now
     @client.update_token!(
@@ -428,6 +436,41 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     expect(@client).to be_expired
   end
 
+  it 'should calculate the expires_at from issued_at when issued_at is set' do
+    expires_in = 3600
+    issued_at = Time.now - expires_in
+    @client.update_token!(
+      :issued_at => issued_at,
+      :expires_in => expires_in
+    )
+    expect(@client.expires_at).to eq issued_at + expires_in
+    expect(@client).to be_expired
+  end
+
+  it 'should calculate expires_at from Time.now when issed_at is NOT set' do
+    expires_in = 3600
+    expires_at = Time.now + expires_in
+    @client.update_token! :expires_in => expires_in
+    expect(@client.expires_at).to be_within(1).of(expires_at)
+    expect(@client).to_not be_expired
+  end
+
+  # This test is to document the way that expires_in has always been used:
+  # If expires_in is set on the client, it always resets the issued_at time
+  # to Time.now
+  it 'sets issued_at to Time.now when expires_in is not set through update_token!' do
+    one_hour = 3600
+    issued_at = Time.now - (2 * one_hour)
+    current_time = Time.now
+
+    @client.issued_at = issued_at
+    @client.expires_in = one_hour
+
+    expect(@client.issued_at).to_not eq issued_at
+    expect(@client.issued_at).to be_within(1).of(current_time)
+    expect(@client).to_not be_expired
+  end
+
   it 'should allow setting expires_at manually' do
     expires_at = Time.now+100
     @client.expires_at = expires_at.to_i
@@ -435,9 +478,50 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     expect(@client).to_not be_expired
   end
 
-  it 'should indicate the token is expired if expired_at nil' do
+  it 'should normalize values of expires_at to instances of time' do
+    time_formats = [DateTime.new, "12:00", 100, Time.new]
+    normalized_time_formats = []
+    time_formats.each do |time|
+      @client.expires_at = time
+      normalized_time_formats << @client.expires_at
+    end
+    normalized_time_formats.each do |time|
+      expect(time).to be_an_instance_of(Time)
+    end
+  end
+
+  it 'should set expires_in when expires_at is set' do
+    issued_at = Time.now
+    expires_at = Time.now+100
+    @client.expires_at = expires_at.to_i
+    @client.issued_at = issued_at
+    expect(@client.expires_in).to be_within(1).of (expires_at - issued_at).to_i
     @client.expires_at = nil
-    expect(@client.expires_within?(60)).to be true
+    expect(@client.expires_in).to be_nil
+  end
+
+  it 'should set expires_in to nil when expires_at is set to nil' do
+    @client.expires_at = nil
+    expect(@client.expires_in).to be_nil
+  end
+
+  it 'should set expires_at when expires_in is set' do
+    expires_in = 100
+    @client.expires_in = expires_in
+    expect(@client.expires_at).to eq (@client.issued_at + expires_in)
+    @client.expires_in = nil
+    expect(@client.expires_at).to be_nil
+  end
+
+  it 'should set expires_at to nil when expires_in is set to nil' do
+    @client.expires_in = nil
+    expect(@client.expires_at).to be_nil
+  end
+
+  it 'should indicate the token is not expired if expired_at nil' do
+    @client.expires_at = nil
+    expect(@client.expires_within?(60)).to be false
+    expect(@client.expired?).to be false
   end
 
   it 'should indicate the token is not expiring when expiry beyond window' do
@@ -471,7 +555,7 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     @client.client_id = 'client-12345'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         [401, {}, 'User authorization failed or something.']
       end
     end
@@ -486,11 +570,11 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     stubs.verify_stubbed_calls
   end
 
-  it 'should raise an error if the token server gives an unexpected status' do
+  it 'should raise a remote server error if the server gives a 5xx status' do
     @client.client_id = 'client-12345'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         [509, {}, 'Rate limit hit or something.']
       end
     end
@@ -501,7 +585,26 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
       @client.fetch_access_token!(
         :connection => connection
       )
-    end).to raise_error(Signet::AuthorizationError)
+    end).to raise_error(Signet::RemoteServerError)
+    stubs.verify_stubbed_calls
+  end
+
+  it 'should raise an unexpected error if the token server gives an unexpected status' do
+    @client.client_id = 'client-12345'
+    @client.client_secret = 'secret-12345'
+    stubs = Faraday::Adapter::Test::Stubs.new do |stub|
+      stub.post('/token') do
+        [309, {}, 'Rate limit hit or something.']
+      end
+    end
+    expect(lambda do
+      connection = Faraday.new(:url => 'https://www.google.com') do |builder|
+        builder.adapter(:test, stubs)
+      end
+      @client.fetch_access_token!(
+        :connection => connection
+      )
+    end).to raise_error(Signet::UnexpectedStatusError)
     stubs.verify_stubbed_calls
   end
 
@@ -511,7 +614,7 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     @client.code = '00000'
     @client.redirect_uri = 'https://www.example.com/'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -537,7 +640,7 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     @client.username = 'johndoe'
     @client.password = 'incognito'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -562,7 +665,7 @@ describe Signet::OAuth2::Client, 'configured for Google userinfo API' do
     @client.client_secret = 'secret-12345'
     @client.refresh_token = '54321'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -748,7 +851,7 @@ JSON
     @client.client_id = 'client-12345'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -789,7 +892,7 @@ JSON
     @client.client_id = 'client-54321'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -824,7 +927,7 @@ JSON
     @client.client_id = 'client-12345'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -854,10 +957,11 @@ JSON
   end
 
   it 'should raise an error if the id token cannot be verified' do
+    pending "Need to set test data"
     @client.client_id = 'client-12345'
     @client.client_secret = 'secret-12345'
     stubs = Faraday::Adapter::Test::Stubs.new do |stub|
-      stub.post('/o/oauth2/token') do
+      stub.post('/token') do
         build_json_response({
           'access_token' => '12345',
           'refresh_token' => '54321',
@@ -897,7 +1001,7 @@ xwIDAQAB
 -----END PUBLIC KEY-----
 PUBKEY
       @client.decoded_id_token(pubkey)
-    end).to raise_error
+    end).to raise_error(JWT::ExpiredSignature)
     stubs.verify_stubbed_calls
   end
 end
@@ -1006,6 +1110,16 @@ describe Signet::OAuth2::Client, 'configured with custom parameters' do
     expect(@client.authorization_uri(:additional_parameters => {'type' => 'new_type', 'new_param' => 'new_val'}).query_values).to eq ({"access_type"=>"offline", "client_id"=>"s6BhdRkqt3", "new_param"=>"new_val",  "response_type"=>"code","redirect_uri"=>"https://example.client.com/callback", "type"=>"new_type"})
   end
 
+  it 'should not have access_type parameter in authorization_uri when we set it to nil in client' do
+    @client.update!(:access_type=>nil)
+    expect(@client.authorization_uri().query_values).to eq ({"client_id"=>"s6BhdRkqt3", "response_type"=>"code", "redirect_uri"=>"https://example.client.com/callback"})
+  end
+
+  it 'should use new access_type parameter as default for authorization_uri' do
+    @client.update!(:access_type=>:online)
+    expect(@client.authorization_uri().query_values).to eq ({"access_type"=>"online", "client_id"=>"s6BhdRkqt3", "response_type"=>"code", "redirect_uri"=>"https://example.client.com/callback"})
+  end
+
   it 'should merge new generate_access_token_request custom parameters' do
     @client.update!(:code=>'12345')
     params = @client.generate_access_token_request(:additional_parameters => {'type' => 'new_type', 'new_param' => 'new_val'})
